import asyncio

import pytest
from fastapi import FastAPI
from httpx import ASGITransport, AsyncClient

from streaq.ui import router
from streaq.ui.deps import get_worker
from streaq.worker import Worker


@pytest.mark.parametrize("prefix", ["", "/streaq"])
async def test_get_pages(worker: Worker, prefix: str):
    app = FastAPI()
    app.include_router(router, prefix=prefix)

    @worker.task()
    async def sleeper(time: int) -> None:
        await asyncio.sleep(time)

    async def _get_worker():
        yield worker

    app.dependency_overrides[get_worker] = _get_worker
    worker.loop.create_task(worker.run_async())

    async with AsyncClient(
        transport=ASGITransport(app=app), base_url="http://test"
    ) as client:
        # queue up some tasks
        tasks = [sleeper.enqueue(i) for i in range(10)]
        for t in tasks[:5]:
            t.delay = 3
        await worker.enqueue_many(tasks)
        # endpoints
        res = await client.get(f"{prefix}/")
        assert res.status_code == 303
        res = await client.get(f"{prefix}/queue")
        assert res.status_code == 200
        res = await client.patch(f"{prefix}/queue")
        assert res.status_code == 200

        short = await sleeper.enqueue(0)
        long = await sleeper.enqueue(5)
        res = await client.get(f"{prefix}/task/{long.id}")
        assert res.status_code == 200
        res = await client.delete(f"{prefix}/task/{long.id}")
        assert res.status_code == 200
        assert res.headers["HX-Redirect"] == f"{prefix}/queue"
        res = await client.get(f"{prefix}/task/{short.id}")
        assert res.status_code == 200



import logging.config
import os
import sys
from multiprocessing import Process
from typing import Annotated, Any, cast

from typer import Exit, Option, Typer
from watchfiles import run_process

from streaq import VERSION, logger
from streaq.utils import default_log_config, import_string
from streaq.worker import Worker

cli = Typer()


def version_callback(value: bool) -> None:
    if value:
        print(f"streaQ v{VERSION}")
        raise Exit()


@cli.command()
def main(
    worker_path: str,
    workers: Annotated[
        int, Option("--workers", "-w", help="Number of worker processes to spin up")
    ] = 1,
    burst: Annotated[
        bool,
        Option(
            "--burst", "-b", help="Whether to shut down worker when the queue is empty"
        ),
    ] = False,
    reload: Annotated[
        bool,
        Option(
            "--reload", "-r", help="Whether to reload the worker upon changes detected"
        ),
    ] = False,
    verbose: Annotated[
        bool,
        Option(
            "--verbose",
            "-v",
            help="Whether to use logging.DEBUG instead of logging.INFO",
        ),
    ] = False,
    version: Annotated[
        bool,
        Option("--version", callback=version_callback, help="Show installed version"),
    ] = False,
    web: Annotated[
        bool,
        Option(
            "--web", help="Run a web UI for monitoring tasks in a separate process."
        ),
    ] = False,
    host: Annotated[
        str, Option("--host", "-h", help="Host for the web UI server.")
    ] = "0.0.0.0",
    port: Annotated[
        int, Option("--port", "-p", help="Port for the web UI server.")
    ] = 8000,
) -> None:
    processes: list[Process] = []
    if web:  # pragma: no cover
        from streaq.ui import run_web

        sys.path.append(os.getcwd())
        worker = cast(Worker[Any], import_string(worker_path))
        logging.config.dictConfig(default_log_config(worker.tz, verbose))
        processes.append(
            Process(
                target=run_web,
                args=(host, port, worker),
            )
        )
    if workers > 1:
        processes.extend(
            [
                Process(
                    target=run_worker,
                    args=(worker_path, burst, reload, verbose),
                )
                for _ in range(workers - 1)
            ]
        )
    for p in processes:
        p.start()
    run_worker(worker_path, burst, reload, verbose)
    for p in processes:
        p.join()


def run_worker(path: str, burst: bool, watch: bool, verbose: bool) -> None:
    """
    Run a worker with the given options.
    """
    if watch:
        run_process(
            ".",
            target=_run_worker,
            args=(path, burst, verbose),
            callback=lambda _: logger.info("changes detected, reloading"),
        )
    else:
        _run_worker(path, burst, verbose)


def _run_worker(path: str, burst: bool, verbose: bool) -> None:
    sys.path.append(os.getcwd())
    worker = cast(Worker[Any], import_string(path))
    logging.config.dictConfig(default_log_config(worker.tz, verbose))
    worker.burst = burst
    worker.run_sync()



PUBLISH_TASK = """
local stream_key = KEYS[1]
local queue_key = KEYS[2]
local task_key = KEYS[3]
local dependents_key = KEYS[4]
local dependencies_key = KEYS[5]
local results_key = KEYS[6]

local task_id = ARGV[1]
local task_data = ARGV[2]
local priority = ARGV[3]
local score = ARGV[4]

if not redis.call('set', task_key, task_data, 'nx') then
  return 0
end

local modified = 0
for i=5, #ARGV do
  local dep_id = ARGV[i]
  if redis.call('exists', results_key .. dep_id) ~= 1 then
    modified = modified + 1
    redis.call('sadd', dependencies_key .. task_id, dep_id)
    redis.call('sadd', dependents_key .. dep_id, task_id)
  end
end

if modified == 0 then
  if score ~= '0' then
    redis.call('zadd', queue_key .. priority, score, task_id)
  else
    return redis.call('xadd', stream_key .. priority, '*', 'task_id', task_id)
  end
end

return 1
"""

FAIL_DEPENDENTS = """
local dependents_key = KEYS[1]
local dependencies_key = KEYS[2]
local task_id = KEYS[3]

local function traverse(task_id, failed)
  if not failed[task_id] then
    failed[task_id] = true
    for _, dep_id in ipairs(redis.call('smembers', dependents_key .. task_id)) do
      traverse(dep_id, failed)
      redis.call('srem', dependencies_key .. dep_id, task_id)
    end
  redis.call('del', dependents_key .. task_id)
  end
end

local visited = {}
traverse(task_id, visited)
local failed = {}
for tid, _ in pairs(visited) do
  if tid ~= task_id then
    table.insert(failed, tid)
  end
end

return failed
"""

UPDATE_DEPENDENTS = """
local dependents_key = KEYS[1]
local dependencies_key = KEYS[2]
local task_id = KEYS[3]

local runnable = {}

local deps = redis.call('smembers', dependents_key .. task_id)
for i = 1, #deps do
  redis.call('srem', dependencies_key .. deps[i], task_id)
  if #redis.call('smembers', dependencies_key .. deps[i]) == 0 then
    table.insert(runnable, deps[i])
  end
end

redis.call('del', dependents_key .. task_id, dependencies_key .. task_id)

return runnable
"""

PUBLISH_DELAYED_TASKS = """
local queue_key = KEYS[1]
local stream_key = KEYS[2]

local current_time = ARGV[1]

for i=2, #ARGV do
  local priority = ARGV[i]
  local queue = queue_key .. priority
  local tids = redis.call('zrange', queue, 0, current_time, 'byscore')
  redis.call('zremrangebyscore', queue, 0, current_time)

  for _, task_id in ipairs(tids) do
    redis.call('xadd', stream_key .. priority, '*', 'task_id', task_id)
  end
end
"""

RECLAIM_IDLE_TASKS = """
local timeout_key = KEYS[1]
local stream_key = KEYS[2]
local group_name = KEYS[3]
local consumer_name = KEYS[4]

local current_time = ARGV[1]
local count = ARGV[2]

local messages = {}

for i=3, #ARGV do
  local priority = ARGV[i]
  local timed_out = redis.call(
    'zrangebyscore',
    timeout_key .. priority,
    0,
    current_time,
    'limit',
    0,
    count
  )

  if #timed_out > 0 then
    messages[priority] = redis.call(
      'xclaim',
      stream_key .. priority,
      group_name,
      consumer_name,
      0,
      unpack(timed_out)
    )
    redis.call('zrem', timeout_key .. priority, unpack(timed_out))
    count = count - #messages[priority]
    if count <= 0 then break end
  end
end

return cjson.encode(messages)
"""

CREATE_GROUPS = """
local stream_key = KEYS[1]
local group_name = KEYS[2]

for i=1, #ARGV do
  local stream = stream_key .. ARGV[i]
  local ok, groups = pcall(redis.call, 'xinfo', 'groups', stream)
  if not ok or #groups == 0 then
    redis.call('xgroup', 'create', stream, group_name, '0', 'mkstream')
  end
end
"""



import asyncio

from streaq import Worker

worker = Worker(redis_url="redis://localhost:6379")


@worker.task()
async def sleeper(time: int) -> int:
    await asyncio.sleep(time)
    return time


@worker.cron("* * * * mon-fri")  # every minute on weekdays
async def cronjob() -> None:
    print("Nobody respects the spammish repetition!")


async def main() -> None:
    async with worker:
        await sleeper.enqueue(3)
        # enqueue returns a task object that can be used to get results/info
        task = await sleeper.enqueue(1).start(delay=3)
        print(await task.info())
        print(await task.result(timeout=5))


if __name__ == "__main__":
    asyncio.run(main())
